#ifndef ECSS_SERVICES_TIMEHELPER_HPP
#define ECSS_SERVICES_TIMEHELPER_HPP

#include <cstdint>
#include <Message.hpp>
#include "TimeAndDate.hpp"

#define SECONDS_PER_MINUTE 60
#define SECONDS_PER_HOUR 3600
#define SECONDS_PER_DAY 86400



/**
 * This class formats the spacecraft time and cooperates closely with the ST[09] time management.
 *
 * The ECSS standard supports two time formats: the CUC and CSD that are described in
 * CCSDS 301.0-B-4 standard. The chosen time format is CDS and it is UTC-based(UTC: Coordinated
 * Universal Time)
 *
 * @note
 * Since this code is UTC-based, the leap second correction must be made. The leap seconds that
 * have been occurred between timestamps should be considered if a critical time-difference is
 * needed
 *
 */
class TimeHelper {
private:
	const uint8_t DaysOfMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

	/**
	 * @param year The year that will be examined if it is a leap year(366 days)
	 * @return if the \p year is a leap year returns true and if it isn't returns false
	 */
	bool IsLeapYear(uint16_t year);

	/**
     * Convert UTC date to elapsed seconds since Unix epoch(1/1/1970 00:00:00).
     *
     * This is a reimplemented mktime() of <ctime> library in an embedded systems way
     *
     * @note
     * This function can convert UTC dates after 1 January 2019 to elapsed seconds since Unix epoch
     *
     * @param TimeInfo the time information/data from the RTC(UTC format)
     * @return the elapsed seconds between a given UTC date(after the Unix epoch) and Unix epoch
     * @todo check if we need to change the epoch to ,the recommended from the standard, 1 January
     * 1958
     */
	uint32_t mkUTCtime(TimeAndDate &TimeInfo);

	/**
     * Convert elapsed seconds since Unix epoch to UTC date.
     *
     * This is a reimplemented gmtime() of <ctime> library in an embedded systems way
     *
     * @note
     * This function can convert elapsed seconds since Unix epoch to UTC dates after 1 January 2019
     *
     * @param seconds elapsed seconds since Unix epoch
     * @return the UTC date based on the \p seconds
     * @todo check if we need to change the epoch to ,the recommended from the standard, 1 January
     * 1958
     */
	TimeAndDate utcTime(uint32_t seconds);

public:

	TimeHelper() = default;

	/**
	 * Generate the CDS time format(3.3 in CCSDS 301.0-B-4 standard).
	 *
	 * The CDS time format consists of two main fields: the time code preamble field(P-field) and
	 * the time specification field(T-field). The P-Field is the metadata for the T-Field. The
	 * T-Field is consisted of two segments: 1)the `DAY` and the 2)`ms of day` segments. The
	 * P-field won't be included in the code, because as the ECSS standards claims, it can be
	 * just implicitly declared.
	 * @param TimeInfo is the data provided from RTC(Real Time Clock)
	 * @return TimeFormat the CDS time format. More specific, 48 bits are used for the  T-field
	 * (16 for the `DAY` and 32 for the `ms of day`)
 	 * @todo time security for critical time operations
 	 * @todo declare the implicit P-field
 	 * @todo check if we need milliseconds
	 */
	static uint64_t generateCDStimeFormat(struct TimeAndDate &TimeInfo);

	/**
	 * Parse the CDS time format(3.3 in CCSDS 301.0-B-4 standard)
	 *
     * @param data time information provided from the ground segment. The length of the data is a
     * fixed size of 48 bits
	 * @return the UTC date that the RTC will be synchronized
	 */
	static TimeAndDate parseCDStimeFormat(const uint8_t *data);

	/**
	 * Dummy function created only to access `mkUTCtime` for testing
	 *
	 * @todo Delete this function
	 */
	uint32_t get_mkUTCtime(TimeAndDate &TimeInfo) {
		return mkUTCtime(TimeInfo);
	}

	/**
	 * Dummy function created only to access `utcTime` for testing
	 *
	 * @todo Delete this function
	 */
	TimeAndDate get_utcTime(uint32_t seconds) {
		return utcTime(seconds);
	}
};

/**
 * Used to access `mkUTCtime` function and `utcTime` function in the static `implementCDSTimeFormat`
 * and in the static `parseCDSTimeFormat` functions
 */
static TimeHelper Access;

#endif //ECSS_SERVICES_TIMEHELPER_HPP
